<?php
/**
 * @file
 *   Contains entity class for entity templates.
 */

/**
 * Template entity class.
 */
class EntityTemplate extends Entity {

  /**
   * The human readable label of the template.
   * @var string
   */
  public $label;

  /**
   * The unique string identifier.
   * @var string
   */
  public $name;

  /**
   * The entity type this template is for.
   * @var string
   */
  public $entity_type;

  /**
   * The bundle of the entity this is for.
   * @var string
   */
  public $bundle;

  /**
   * The parameters needed to create an entity from this template.
   * @var Array
   */
  public $parameters = array();

  /**
   * The template.
   * @var string
   */
  public $template;

  /**
   * Get the parameters.
   *
   * Use this to add the site settings parameter info to all templates.
   */
  function getParameters() {
    return $this->parameters + array(
      'site' => array(
        'type' => 'site',
        'label' => t('Site Information'),
        'description' => t('Site-wide settings and other global information.'),
        'property info alter' => array('RulesData', 'addSiteMetadata'),
        'property info' => array(),
        'optional' => TRUE,
      ),
      'self' => array(
        'label' => t('Entity being Created'),
        'description' => t('The entity being created by this template.'),
        'type' => $this->entity_type,
        'bundle' => $this->bundle,
        'optional' => TRUE,
      )
    );
  }

  /**
   * Build the template.
   */
  function buildTemplate() {
    if (empty($this->template)) {
      return $this->prepareBlankTemplate();
    }

    // Build the template entity. This is almost the same as calling 'import'
    // but we pass the template in to the create method as well.
    $vars = drupal_json_decode($this->template);
    if (is_array($vars)) {
      $vars['__template'] = $this;
      $entity = entity_get_controller($this->entity_type)->create($vars);
    }

    if (empty($entity)) {
      drupal_set_message(t('Could not create entity from template.'), 'warning');
      return $this->prepareBlankTemplate();
    }

    // Find field collection items and tweak the entity array.
    // Move this into a hook/make more generically pluggable.
    foreach ($entity as $prop_name => &$prop) {
      if (($field = field_info_field($prop_name)) && $field['type'] == 'field_collection') {
        foreach ($prop as $langcode => &$items) {
          if (!is_array($items)) {
            continue;
          }

          foreach ($items as $delta => &$item) {
            if (!empty($item['entity']) && is_array($item['entity'])) {
              $item['entity'] = entity_create('field_collection_item', $item['entity']);
            }
          }
        }
      }
    }

    return $entity;
  }

  /**
   * Get a nested property.
   */
  function getNestedWrapper($wrapped_params, $selector) {
    $parts = explode(':', $selector);
    $namespace = str_replace('-', '_', array_shift($parts));
    $selected = $wrapped_params[$namespace];

    if (empty($selected)) {
      return FALSE;
    }

    try {
      while (($sub = array_shift($parts)) !== NULL) {
        try {
          $sub = str_replace('-', '_', $sub);
          $selected = $selected->get($sub);
        }
        catch (EntityMetadataWrapperException $e) {
          return FALSE;
        }
      }
    }
    catch (Exception $e) {
      return FALSE;
    }

    return $selected;
  }

  /**
   * Apply a data selector to a property.
   *
   * @param EntityMetadataWrapper $wrapper
   *   The wrapper of property to set.
   * @param $selector
   *   The selector to find the data value.
   * @param $params
   *   The wrapper parameters.
   * @param RulesDataProcessor $processors
   *   Rules data processor to tweak values/
   */
  function applyDataSelector($wrapper, $selector, $wrapped_params, $processor = NULL) {
    // Special behaviour for lists.
    // For lists we allow the use of a comma separated list of properties
    // these properties can be lists in their own right or individual
    // instances of whatever the list contains.
    if ($wrapper instanceof EntityListWrapper) {
      $selectors = explode(',', $selector);
      foreach ($selectors as $sel) {
        if ($selected = $this->getNestedWrapper($wrapped_params, $sel)) {
          if ($selected instanceof EntityListWrapper) {
            // Selected is a list, so loop over what we have and add them to
            // the property.
            foreach ($selected as $item) {
              try {
                if ($item->value()) {
                  $wrapper[] = $item->value();
                }
              }
              catch (EntityMetadataWrapperException $e) {}
            }
          }
          else {
            // Get the value from the selected wrapper. We wrap this in a try
            // block to make sure that the system doesn't completely fail when
            // one of the parent values is missing.
            try {
              if (is_callable(array($selected, 'value')) && $selected->value()) {
                $wrapper[] = $selected->value();
              }
            }
            catch (EntityMetadataWrapperException $e) {}
          }
        }
      }
    }
    else {
      $selected = $this->getNestedWrapper($wrapped_params, $selector);
      try {
        if (is_object($selected) && $selected->value()) {
          $value = $selected->value();
          if (!empty($processor)) {
            $state = new RulesState();
            foreach ($wrapped_params as $name => $param_wrapper) {
              if ($name == 'site') {
                continue;
              }
              $state->addVariable($name, $param_wrapper->value(), $param_wrapper->info());
            }

            $value = $processor->process($value, $wrapper->info(), $state, new RulesAction('entity_template_create'));
          }
          $wrapper->set($value);
        }
      }
      catch (EntityMetadataWrapperException $e) {}
    }
  }

  /**
   * Create new entity from template.
   */
  function createNewEntityFromTemplate($parameters) {
    $entity = $this->buildTemplate();
    $wrapper = entity_metadata_wrapper($this->entity_type, $entity);

    // Make sure params are wrapped.
    $wrapped_params = array();
    foreach ($parameters as $namespace => $param) {
      if ($param instanceof EntityMetadataWrapper) {
        $wrapped_params[$namespace] = $param;
      }
      else {
        $info = $this->parameters[$namespace];
        $wrapped_params[$namespace] = entity_metadata_wrapper($info['type'], $param);
      }
    }

    // Add Site Params.
    if (empty($wrapped_params['site'])) {
      $wrapped_params['site'] = entity_metadata_wrapper('site', FALSE, $this->getParameters()['site']);
    }

    // Add Self Param.
    if (empty($wrapped_params['self'])) {
      $wrapped_params['self'] = entity_metadata_wrapper($this->entity_type, $entity, $this->getParameters()['self']);
    }

    // Loop over properties and convert data selections.
    $properties = $wrapper->getPropertyInfo();
    foreach ($entity as $prop => $value) {
      if (!is_array($value) || empty($properties[$prop])) {
        continue;
      }

      // Set the value to null.
      $entity->{$prop} = NULL;

      // Work out what method to use.
      $method = !empty($value['select']) ? 'select' : 'direct';
      if (!empty($value['method'])) {
        $method = $value['method'];
      }

      $property = $properties[$prop];
      if ($method == 'sub_select') {
        if (!is_array(reset($value['sub_select']))) {
          // If the first item isn't an array then we are dealing with a pre
          // update template. Include this code for backwards compatibility.
          foreach (array_filter($value['sub_select']) as $sub_prop => $selector) {
            // As options are stored in the same way as sub selectors we check
            // that the sub property exists.
            if (empty($property['property info'][$sub_prop])) {
              continue;
            }

            // Apply the selector.
            $processors = NULL;
            if (!empty($value['sub_select'][$sub_prop.'_template_options'])) {
              $processors = $value['sub_select'][$sub_prop.'_template_options'];
              RulesDataProcessor::prepareSetting(
                $processors,
                $property['property info'][$sub_prop],
                $this->getParameters()
              );
            }
            $prop_wrapper = $wrapper->{$prop};
            if ($prop_wrapper instanceof EntityListWrapper) {
              $prop_wrapper = $prop_wrapper[0];
            }

            $this->applyDataSelector($prop_wrapper->{$sub_prop}, $selector, $wrapped_params, $processors);
          }
        }
        else if ($list_type = entity_property_list_extract_type($property['type'])) {
          // This is a list so we apply over and over again for all the items.
          $sub_select_items = $value['sub_select'];
          if (!is_numeric(key($value['sub_select']))) {
            $sub_select_items = array($value['sub_select']);
          }

          foreach ($sub_select_items as $sub_delta => $sub_info) {
            // Prepare the wrapper.
            // Special case for field collections.
            if ($list_type == 'field_collection_item') {
              $item = &$entity->{$prop}[LANGUAGE_NONE][$sub_delta];
              $field_collection_item = field_collection_field_get_entity($item, $prop);
              $prop_wrapper = entity_metadata_wrapper('field_collection_item', $field_collection_item);
            }
            else {
              $prop_wrapper = $wrapper->{$prop}[$sub_delta];
            }

            // Get sub property info.
            $sub_properties = $prop_wrapper->getPropertyInfo();
            foreach ($sub_info as $sub_prop => $options) {
              if (empty($sub_properties[$sub_prop])) {
                continue;
              }

              if (!empty($options['parameter']['select'])) {
                // Prepare the processors.
                $processors = NULL;
                if (!empty($options['parameter']['select_template_options'])) {
                  $processors = $options['parameter']['select_template_options'];
                  RulesDataProcessor::prepareSetting(
                    $processors,
                    $sub_properties[$sub_prop],
                    $this->getParameters()
                  );
                }
                $this->applyDataSelector($prop_wrapper->{$sub_prop}, $options['parameter']['select'], $wrapped_params, $processors);
              }
              else if (!empty($options['direct_input'])) {
                if ($prop_wrapper->{$sub_prop} instanceof EntityDrupalWrapper || $prop_wrapper->{$sub_prop} instanceof EntityValueWrapper) {
                  if (!empty($options['direct_input']['element'])) {
                    $prop_wrapper->{$sub_prop} = $options['direct_input']['element'];
                  }
                }
                else {
                  $prop_wrapper->{$sub_prop} = $options['direct_input'];
                }
              }
            }
          }
        }
        else {
          // Single item with sub properties.
          $prop_wrapper = $wrapper->{$prop};
          $sub_properties = $prop_wrapper->getPropertyInfo();

          foreach ($value['sub_select'] as $sub_prop => $options) {
            // @todo: Can we move this to a method?
            if (empty($sub_properties[$sub_prop])) {
              continue;
            }

            if (!empty($options['parameter']['select'])) {
              // Prepare the processors.
              $processors = NULL;
              if (!empty($options['parameter']['select_template_options'])) {
                $processors = $options['parameter']['select_template_options'];
                RulesDataProcessor::prepareSetting(
                  $processors,
                  $sub_properties[$sub_prop],
                  $this->getParameters()
                );
              }
              $this->applyDataSelector($prop_wrapper->{$sub_prop}, $options['parameter']['select'], $wrapped_params, $processors);
            }
            else if (!empty($options['direct_input'])) {
              if ($prop_wrapper->{$sub_prop} instanceof EntityDrupalWrapper || $prop_wrapper->{$sub_prop} instanceof EntityValueWrapper) {
                if (!empty($options['direct_input']['element'])) {
                  $prop_wrapper->{$sub_prop} = $options['direct_input']['element'];
                }
              }
              else {
                $prop_wrapper->{$sub_prop} = $options['direct_input'];
              }
            }
          }
        }
      }
      else if ($method == 'select') {
        // Apply the selector.
        $processors = NULL;
        if (!empty($value['select_template_options'])) {
          $processors = $value['select_template_options'];
          RulesDataProcessor::prepareSetting(
            $processors,
            $property,
            $this->getParameters()
          );
        }
        $selector = $value['select'];
        $this->applyDataSelector($wrapper->{$prop}, $selector, $wrapped_params, $processors);
      }
      else if ($method == 'direct' && empty($property['field'])) {
        $entity->{$prop} = $value['value'];
      }
      else if ($method == 'direct') {
        $entity->{$prop}[LANGUAGE_NONE] = !empty($value[LANGUAGE_NONE]) ? $value[LANGUAGE_NONE] : array();
      }

      // Get rid of empty field collection items.
      if (
        in_array($property['type'], array('field_collection_item', 'list<field_collection_item>')) &&
        is_array($entity->{$prop}[LANGUAGE_NONE])
      ) {
        foreach ($entity->{$prop}[LANGUAGE_NONE] as $fc_delta => $fc_item) {
          $fc_entity = field_collection_field_get_entity($fc_item);
          if (field_collection_item_is_empty($fc_entity)) {
            unset($entity->{$prop}[LANGUAGE_NONE][$fc_delta]);
          }
        }
        $entity->{$prop}[LANGUAGE_NONE] = array_values($entity->{$prop}[LANGUAGE_NONE]);
      }
    }

    // Perform ctools style substitutions.
    $parameters['self'] = $entity;
    foreach ($entity as $prop => &$value) {
      if (is_string($value)) {
        $value = $this->replaceCtoolsSubstitutions($value, $parameters);
      }
      else if (($field_info = field_info_field($prop)) && is_array($value)) {
        foreach ($value as $langcode => &$langitem) {
          if (!is_array($langitem) && !($langitem instanceof Iterator)) {
            continue;
          }

          foreach ($langitem as $delta => &$item) {
            if (!is_numeric($delta)) {
              continue;
            }

            // Special handling for field collections.
            // We turn the 'entity' array back into a field collection item
            // object.
            if ($field_info['type'] == 'field_collection') {
              if (is_array($item['entity'])) {
                $item['entity'] = entity_create('field_collection_item', $item['entity']);
              }

              // Do tokens on string fields on the field collection item.
              foreach ($item['entity'] as &$item_prop) {
                if (!is_array($item_prop)) {
                  continue;
                }

                foreach ($item_prop as $item_langcode => &$item_langitem) {
                  if (!is_array($item_langitem)) {
                    continue;
                  }

                  foreach ($item_langitem as $item_delta => &$item_item) {
                    if (!is_numeric($item_delta)) {
                      continue;
                    }

                    foreach ($item_item as $item_key => &$item_val) {
                      if (is_string($item_val)) {
                        $item_val = $this->replaceCtoolsSubstitutions($item_val, $parameters);
                      }
                    }
                  }
                }
              }
            }

            foreach ($item as $key => &$val) {
              if (is_string($val)) {
                $val = $this->replaceCtoolsSubstitutions($val, $parameters);
              }
            }
          }
        }
      }
    }

    // Finaly tell the entity which template made it.
    $entity->entity_template = $this->name;

    module_invoke_all('entity_template_create_entity', $this, $entity, $this->entity_type);
    rules_invoke_event('entity_template_create_entity', $this, entity_metadata_wrapper($this->entity_type, $entity));

    return $entity;
  }

  /**
   * Prepare a blank template.
   */
  function prepareBlankTemplate() {
    $info = entity_get_info($this->entity_type);
    $values = array();
    if (!empty($info['entity keys']['bundle'])) {
      $values[$info['entity keys']['bundle']] = $this->bundle;
    }
    $entity = entity_create($this->entity_type, $values);
    module_invoke_all('entity_template_prepare_blank_template', $this, $entity, $this->entity_type);
    return $entity;
  }

  /**
   * Get an array of ctools context for the flexiform.
   *
   * @param array $entities
   *   An array of entities keyed by namespace. If NULL, placeholder contexts
   *   will be used.
   *
   * @return ctools_context[]
   *   An array of ctools contexts.
   */
  public function getCtoolsContexts($parameters = array()) {
    global $user;

    ctools_include('context');
    $contexts = array(
      'site' => ctools_context_create('token'),
      'current-user' => ctools_context_create('entity:user', $user),
      'self' => !empty($parameters['self']) ?
        ctools_context_create('entity:' . $this->entity_type, $parameters['self']) :
        ctools_context_create_empty('entity:' . $this->entity_type)
    );
    $contexts['site']->keyword = 'global';
    $contexts['current-user']->keyword = 'current-user';
    $contexts['current-user']->identifier = t('Logged-in user');
    $contexts['self']->keyword = 'self';

    if (is_array($this->parameters)) {
      $available_context_plugins = ctools_get_contexts();
      foreach ($this->parameters as $namespace => $info) {
        if (entity_get_info($info['type'])) {
          $type = 'entity:' . $info['type'];
        }
        else if ($info['type'] == 'text') {
          $type = 'string';
        }
        else if (!empty($available_context_plugins[$info['type']])) {
          $type = $info['type'];
        }
        else {
          continue;
        }

        if (!empty($parameters[$namespace])) {
          $contexts[$namespace] = ctools_context_create($type, $parameters[$namespace]);
        }
        else {
          $contexts[$namespace] = ctools_context_create_empty($type);
        }
        $contexts[$namespace]->keyword = $namespace;
        $contexts[$namespace]->identifier = $info['label'];
      }
    }

    return $contexts;
  }

  /**
   * Replace ctools substitutions with their values.
   *
   * To allow conventional ctools tokens to work in the actual entity that gets
   * created we use a slightly different syntax. Tokens start with @: instead
   * of %.
   *
   * @param string $string
   *   The string we want to replace in.
   * @param array $parameters
   *   An array of entities keyed by namespace. If NULL, placeholder contexts
   *   will be used.
   * @param array $keywords
   *   Optionally provide additional keywords to replace.
   *
   * @return string
   *   $string with it's substitutions replaced.
   */
  public function replaceCtoolsSubstitutions($string, $parameters, $keywords = array()) {
    $contexts = $this->getCtoolsContexts($parameters);

    // Convert normal ctools tokens into ignorable ctools tokens.
    $string = str_replace('%', '@|', $string);
    $string = str_replace('@:', '%', $string);
    $string = ctools_context_keyword_substitute($string, $keywords, $contexts, array('sanitize' => FALSE));
    $string = str_replace('@|', '%', $string);
    return $string;
  }

  /**
   * Build a list of possible ctools substitutions.
   *
   * @param array $keywords
   *   Optionally provide additional keywords to show.
   *
   * @return array
   *   A render array of substitutions.
   */
  public function getCtoolsSubstitutionsList($keywords = array()) {
    $content = array(
      '#theme' => 'table',
      '#header' => array(t('Keyword'), t('Value')),
      '#rows' => array(),
    );

    foreach ($this->getCtoolsContexts() as $context) {
      foreach (ctools_context_get_converters('@:' . check_plain($context->keyword) . ':', $context) as $keyword => $title) {
        $content['#rows'][] = array(
          check_plain($keyword),
          t('@identifier: @title', array('@title' => $title, '@identifier' => $context->identifier)),
        );
      }
    }

    if (count($content['#rows'])) {
      return $content;
    }
    else {
      return array();
    }
  }
}

/**
 * Entity Controller
 */
class EntityTemplateController extends EntityAPIControllerExportable {

  /**
   * {@inheritdoc}
   */
  public function export($entity, $prefix = '') {
    $tmpEntity = clone $entity;
    $tmpEntity->template = drupal_json_decode($tmpEntity->template);

    $vars = get_object_vars($tmpEntity);
    unset($vars[$this->statusKey], $vars[$this->moduleKey], $vars['is_new']);
    if ($this->nameKey != $this->idKey) {
      unset($vars[$this->idKey]);
    }

    return static::exportVarJsonExport($vars, $prefix);
  }

  /**
   * Process export json.
   *
   * @param $var
   *   The variable to be json encoded.
   * @param $prefix
   *   The prefix.
   *
   * @return string
   *   A portion of encoded json.
   */
  public static function exportVarJsonExport($var, $prefix = '') {
    if (is_array($var) && $var) {
      // Defines whether we use a JSON array or object.
      $use_array = ($var == array_values($var));
      $output = $use_array ? "[" : "{";

      foreach ($var as $key => $value) {
        if ($use_array) {
          $values[] = static::exportVarJsonExport($value, '  ');
        }
        else {
          $values[] = static::exportVarJsonExport((string) $key, '  ') . ' : ' . static::exportVarJsonExport($value, '  ');
        }
      }
      // Use several lines for long content. However for objects with a single
      // entry keep the key in the first line.
      if (strlen($content = implode(', ', $values)) > 70 && ($use_array || count($values) > 1)) {
        $output .= "\n  " . implode(",\n  ", $values) . "\n";
      }
      elseif (strpos($content, "\n") !== FALSE) {
        $output .= " " . $content . "\n";
      }
      else {
        $output .= " " . $content . ' ';
      }
      $output .= $use_array ? ']' : '}';
    }
    elseif (is_string($var)) {
      if (stripos($var, "\n") !== FALSE) {
        $var = str_replace("\r\n", "\n", $var, $count);
        $new_var = array(
          'FN::Implode' => explode("\n", $var),
          'FN::Implode::Separator' => $count ? '\r\n' : '\n',
        );
        $output = static::exportVarJsonExport($new_var, '  ');
      }
      else {
        $output = json_encode($var, JSON_HEX_QUOT);
      }
    }
    else {
      $output = drupal_json_encode($var);
    }

    if ($prefix) {
      $output = str_replace("\n", "\n$prefix", $output);
    }

    return $output;
  }

  /**
   * Process json export being imported.
   *
   * @param $var
   *   The decoded values.
   *
   * @return mixed
   *   Processed values.
   */
  public static function importVarJsonImport($var) {
    if (is_array($var)) {
      foreach ($var as $key => $value) {
        if ($key === 'FN::Implode') {
          return implode(
            str_replace(array('\r\n', '\n'), array("\r\n", "\n"), $var['FN::Implode::Separator'] ?? '\n'),
            $value
          );
        }
        else {
          $var[$key] = static::importVarJsonImport($value);
        }
      }
    }

    return $var;
  }

  /**
   * {@inheritdoc}
   */
  public function import($export) {
    $vars = static::importVarJsonImport(drupal_json_decode($export));

    if (is_array($vars)) {
      $entity = $this->create($vars);
    }
    else {
      return FALSE;
    }

    if (!is_string($entity->template)) {
      $entity->template = drupal_json_encode($entity->template);
    }

    unset($entity->entity_template_tags);
    return $entity;
  }

  /**
   * {@inheritdoc}
   */
  public function invoke($hook, $entity) {
    if (in_array($hook, array('insert', 'update'))) {
      $this->storeTags($entity);
    }

    if ($hook == 'update') {
      $this->purgeRevisions($entity);
    }

    parent::invoke($hook, $entity);

    // Invoke rules.
    if (module_exists('rules')) {
      rules_invoke_event($this->entityType . '_' . $hook, $entity);
    }
  }

  /**
   * Purge older revisions from the object.
   */
  public function purgeRevisions($template) {
    if ($max_historical_revisions = variable_get('entity_template_max_revisions', NULL)) {
      $vids = db_select('entity_template_revision', 'r')
        ->fields('r', array('vid'))
        ->condition('r.id', $template->id)
        ->condition('r.vid', $template->vid, '<>')
        ->orderBy('r.timestamp', 'DESC')
        ->orderBy('r.vid', 'DESC')
        ->range($max_historical_revisions, 1000)
        ->execute()
        ->fetchCol();

      foreach ($vids as $vid) {
        entity_revision_delete('entity_template', $vid);
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function attachLoad(&$queried_entities, $revision_id = FALSE) {
    parent::attachLoad($queried_entities, $revision_id);

    $query = db_select('entity_template_tags', 't')
      ->fields('t', array('id', 'tag'));
    $or = db_or();
    foreach ($queried_entities as $id => $queried_entity) {
      $or->condition(db_and()
        ->condition('t.id', $queried_entity->id)
        ->condition('t.vid', $queried_entity->vid)
      );
    }
    $query->condition($or);
    foreach ($query->execute() as $row) {
      $tags[$row->id][] = $row->tag;
    }

    foreach ($queried_entities as &$queried_entity) {
      $queried_entity->tags = isset($tags[$queried_entity->id]) ? $tags[$queried_entity->id] : array();
    }
  }

  /**
   * Store the tags.
   */
  public function storeTags($entity) {
    db_delete('entity_template_tags')
      ->condition('id', $entity->id)
      ->condition('vid', $entity->vid)
      ->execute();
    if (!empty($entity->tags)) {
      $query = db_insert('entity_template_tags')->fields(array('id', 'vid', 'tag'));
      foreach ($entity->tags as $tag) {
        $query->values(array(
          'id' => $entity->id,
          'vid' => $entity->vid,
          'tag' => $tag,
        ));
      }
      $query->execute();
    }
  }

}
